import errno
import os
from copy import deepcopy
from pathlib import Path
from .Argparser import Argparser
from .utils.Type_Convertor import Type_Convertor
from .utils.Flag import Flag

class Configer(object):
    '''
        The Configer attemtp to make a light-weight solution for configurating your program, 
        which offer a simple syntax for declare the arguments in the configure file (.ini suffix).
        Moreover, I try to implement a highly customer reader, which allow the user to declare 
        the instance of customer class with registered their constructor simply. 
        
        Hope such trivial contribution will let your work become easier ~ ~ God bless you.
    '''
    def __init__(self, description:str = "", cmd_args:bool = False, split_chr:str = " = "):
        '''
            description (option) : 
                A customer helper information which describe the functionality of your configer file.
            
            cmd_args (option) :
                Allow the commend line argument update the declared value of configer file.
                
            declare_split_chr (option) :
                The character make use to split the declaration string in configer file.
                For example, 'a*13@int' which means the argument 'a' contain interger value 13,
                and the '*' is the declare_split_chr.
        '''
        self.__doc_str = "Description : \n" + description
        self.__typ_cnvt = Type_Convertor()
        self.__split_chr = split_chr
        self.__flag = Flag().FLAGS
        self.__cmd_args = cmd_args
            
    # The cell-base Intereactive Enviroment Support function
    def cfg_from_str(self, raw_cfg_text:str):
        ''' 
            raw_cfg_text :
                The string which declare the arguments with the same syntax used in config file. 
        '''
        self.__cfg_parser(raw_cfg_text)
        # build the flag object 
        self.__flag.__dict__ = self.__dict__
         
    def cfg_from_ini(self, cfg_path:str):
        '''
            cfg_path :
                The path which locate the '*.ini' config file.
        '''
        def chk_src(cfg_path):
            if not cfg_path.exists():
                raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), str(cfg_path))
            if not cfg_path.suffix == ".ini":
                raise ValueError("The file extension should be 'ini', instead of '{}'".format(cfg_path.suffix))
            
        try:
            # check config path validation
            cfg_path = Path(cfg_path)
            chk_src(cfg_path)

            with cfg_path.open('r') as cfg_ptr: 
                raw_cfg_text = cfg_ptr.read()
            
        except FileNotFoundError as fnf_err:
            print(fnf_err) ; raise
        except ValueError as val_err:
            print(val_err) ; raise
        except PermissionError as per_err:
            print(per_err) ; raise
        except Exception as ex:
            print(ex) ; raise
        else:
            self.__cfg_parser(raw_cfg_text)
        
        # build the flag object 
        self.__flag.__dict__ = self.__dict__

    # return an absl style flag to store all of the args.
    def get_cfg_flag(self):
        return self.__flag

    def get_doc_str(self):
        return self.__doc_str

    # merge conf suppose 2 config have overlap section, otherwise use 'concate' method!
    def merge_conf(self, cfg, override=True):
        
        def hier_merge(sf_dict, cfg_dict):
            for sec_key, sec_val in cfg_dict.items():
                # same section exists
                if sec_key in sf_dict.keys():
                    if not override:
                        raise RuntimeError(f"Key '{sec_key}' in input config already exists in merged config!!")
                    
                    # if both self-dict and cfg_dict are dict, merge it hierachically!
                    if isinstance(sec_val, dict) and isinstance(sf_dict[sec_key], dict):
                        hier_merge(sf_dict[sec_key], sec_val)
                    else:
                        sf_dict[sec_key] = sec_val

                # directly feed new val
                else:
                    sf_dict[sec_key] = sec_val

        # prevent checking private vars
        cfg_dict = { k:v for k, v in cfg.__dict__.copy().items() \
                                    if '_' != k[0] }
        hier_merge(self.__dict__, cfg_dict)
    
    # concate just means merge2conf "without" any override!!
    def concate_cfg(self, cfg):
        cp_cfg = deepcopy(self)
        cp_cfg.merge_conf(cfg, override=False)
        return cp_cfg

    ## Configuration operator support :
    #  all of operator will be forced to return value!!
    # 1. merge operator, force to override!
    def __or__(self, cfg):
        cp_cfg = deepcopy(self)
        cp_cfg.merge_conf(cfg, override=True)
        return cp_cfg

    # 2. concate operator 
    def __add__(self, cfg):
        return self.concate_cfg(cfg, override=False)


    # utils of config parser
    def __preproc_cfgstr(self, cfg_str:str) -> str:
        # eliminate the line full of white-space without any text
        cfg_str = cfg_str.strip() 
        # skip empty line and comment line 
        if len(cfg_str) == 0 or cfg_str[0] == '#': 
            return ''
        # strip inline comment's content with strip extra-whitespace
        return cfg_str.split('#')[0].strip()

    def __get_sec(self, cfg_str:str) -> str:
        if '[' != cfg_str[0]:
            return ''
        # more robust with Indented section
        beg, end = cfg_str.find('['), cfg_str.rfind(']')
        if end == -1:
            raise RuntimeError("Configuration Error : Invalid section notation, missing ']' at end of line")
        sec_key_str = cfg_str[beg+1 : end].strip()
        return sec_key_str
    
    def __idx_sec_by_dot(self, sec_keys_str:str) -> [dict, str]:
        sec_name_lst = sec_keys_str.split('.')
        # Before easy_configer 1.3.4 ver, all section is builded upon this level
        if len(sec_name_lst) == 1:
            return self.__dict__, sec_keys_str

        ## Support toml like 'hierachical' format!!
        #  dynamically search the hierachical section begin from the 'next layer' of self.__dict__
        idx_sec = self.__dict__[ sec_name_lst.pop(0) ]
        #  keep the index point to the node "parent", since the child node will be init as dict!
        try:
            for sec in sec_name_lst[:-1]:
                idx_sec = idx_sec[sec]
        except:
            raise RuntimeError(f"The parent node of {sec} is not defined yet, it's invalid for directly made the child node")

        return idx_sec, sec_name_lst[-1]

    def __get_declr_dict(self, cfg_str:str) -> dict : 
        if self.__split_chr not in cfg_str:
            raise RuntimeError(f"Configuration Error : Split character '{self.__split_chr}' not found in '{cfg_str}'")
        
        try:
            var_name, val_str = cfg_str.split(self.__split_chr)
            # support '$' args interpolation
            var_val = self.__dict__[ val_str[1:] ]  if val_str[0] == '$' \
                        else self.__typ_cnvt.convert(val_str)
        except:
            raise RuntimeError(f"Configuration Error : Invalid config string ' {cfg_str}' ")

        return { var_name : var_val }

    def __cfg_parser(self, raw_cfg_text:str):
        '''
            raw_cfg_text :
                The string which declare the arguments with the same syntax used in config file.
        '''
        cur_sec_keys = ''
        for lin in raw_cfg_text.splitlines():
            # strip empty space and 'skip' empty line in cfgstr
            cfg_str = self.__preproc_cfgstr(lin)
            if not cfg_str:
                continue

            # get the section key of config string (if there exists)
            sec_keys_str = self.__get_sec(cfg_str)
            if sec_keys_str:
                idx_sec, idx_sec_key = self.__idx_sec_by_dot(sec_keys_str)
                if idx_sec_key in idx_sec.keys():
                    raise RuntimeError(f'Re-defined config, {sec_keys_str} section will be overrided!!')
                idx_sec[idx_sec_key] = {}
                cur_sec_keys = sec_keys_str
                
            # parse variable assignment string
            else:
                val_dict = self.__get_declr_dict(cfg_str)
                if cur_sec_keys != '':
                    idx_sec, idx_sec_key = self.__idx_sec_by_dot(cur_sec_keys)
                    idx_sec[idx_sec_key].update( val_dict )
                else:
                    self.__dict__.update( val_dict )

        # Update the namespace value via commend-line input 
        if self.__cmd_args:
            Argparser.args_from_cmd(self.__idx_sec_by_dot)

    # Display the namespace which record all of the declared arguments
    #   for the inner-node structure, iter-call __str__ wrapper !!
    def __str__(self):
        key_str = [ str(key) for key in self.__dict__.keys() if key[0] != '_' ] 
        return "Namespace : \n" + ", ".join(key_str)

    # For the declare the instance of user customized class 
    def regist_cnvtor(self, *args:list, **kwargs:dict):
        self.__typ_cnvt.regist_cnvtor(*args, **kwargs)
    
    # show split character
    @property
    def split_char(self):
        return self.__split_chr



from functools import partial
import ast
import re

import warnings
def warning_on_one_line(message, category, filename, lineno, file=None, line=None):
    return "{}:{}: {}: {}\n".format(filename, lineno, category.__name__, message)
warnings.formatwarning = warning_on_one_line


class Type_Convertor(object):
    '''
        As the name implies, this helper class of Configer will convert the raw string of config file
        into variable of the corresponding type.
        
        Now, i have update the security policy to prevent loading the harmful code to your system ~
        feel free to use the default type converter. 

        However, we still apply eval in regist_cnvtor, plz be careful about register the customized classes ~
    '''
    def __init__(self, typ_split_chr:str = '@'):
        '''
            typ_split_chr (option) :
                The character make use to split the declaration string in configer file.
                For example, 'a*13@int' which means the argument 'a' is interger type data,
                and the '@' is the typ_split_chr.
        '''
        self.__split_chr = typ_split_chr

        # fundemental api-string supported in built-in 
        # i.e. "{{{}}}".format("'key':56")  ->  "{'key':56}", ast.literal_eval("{'key':56}") -> {'key':56}
        #      "{{{}}}".format("1, 2, 3")  ->  "{1, 2, 3}", ast.literal_eval("{1, 2, 3}") -> {1, 2, 3}
        #      so, dict and set apply the same api_str called curly_braces_str!
        curly_braces_str = "{{{}}}" ; lst_str = "[{}]" ; tuple_str = "({})" 
        # add security policy : we applied ast.literal_eval to eval str
        cnt_wrap = lambda val, api_str : ast.literal_eval(api_str.format(val))
       
        # create various converter by wrapping api_str into ast.literal_eval :
        curly_braces_cnvt = partial(cnt_wrap, api_str=curly_braces_str)
        tuple_cnvt = partial(cnt_wrap, api_str=tuple_str)
        lst_cnvt = partial(cnt_wrap, api_str=lst_str)

        # deal with bool(.) constructor feature, empty str regard as False
        bool_cnvt = lambda val: False if val == 'False' else bool(val)
        
        # basic datatype, build-in container, collection
        self.__default_cnvtor = {"str":str, "int":int, "float":float, "bool":bool_cnvt,
                                    "dict":curly_braces_cnvt, "set":curly_braces_cnvt, "tuple":tuple_cnvt, "list":lst_cnvt}
        self.__customized_cnvtor = {}

    def convert(self, cfg_raw_str:str):
        '''
            cfg_raw_str :
                The string which declare the arguments with the same syntax used in config file.
        '''    
        # force to add split character at the end of parsed string
        cfg_raw_str = cfg_raw_str + self.__split_chr if (not self.__split_chr in cfg_raw_str) else cfg_raw_str
        try:
            val_str, typ = cfg_raw_str.split(self.__split_chr)
        except:
            raise RuntimeError
        
        if typ in self.__customized_cnvtor.keys():
            args = ast.literal_eval(val_str)
            # For init customized class, we provide args, kwargs or default init  
            if isinstance(args, dict):
                return self.__customized_cnvtor[typ](**args)
            elif isinstance(args, list):
                return self.__customized_cnvtor[typ](*args)
            else:  # invalid type of args is considered as default init  
                warnings.warn(f"You're initialized class '{typ}' with default arguments!")
                return self.__customized_cnvtor[typ]()
        
        # support 'None' placeholder and [], {} eval, instead of define '@type'
        elif (val_str == 'None') or (not typ):  
            return ast.literal_eval(val_str)
        
        # type-validator : we use ast.literal_eval and it need to \
        else:  # strip '[', ']', '{', '}' notation before feeding into 'default' type-conveter
            stripped_val_str = re.sub(r"[\[\]\{\}]", "", val_str)
        return self.__default_cnvtor[typ](stripped_val_str)
    
    def regist_cnvtor(self, type_name:str = None, cnvt_func:callable = None):
        '''
            type_name :
                Name of registered function, namely the name of customized class.
            
            cnvt_func :
                The function for converting the string object into the customized class instance.
        '''
        assert callable(cnvt_func), "The converter function should be callable."
        assert isinstance(type_name, str) and len(type_name) > 0, "The cnvt_name should be given"
        
        func_wrap = lambda *args, **kwargs : cnvt_func(*args, **kwargs)
        self.__customized_cnvtor[type_name] = func_wrap


from .Configer import Configer

from argparse import ArgumentParser
from omegaconf import OmegaConf
import yaml

class IO_Converter(object):
    def __init__(self):
        # support package regist :
        self.dispatcher = {
            'argparse' : self.argparse_cnvt,
            'omegacfg' : self.omegacfg_cvnt,
            'yaml' : self.yaml_cnvt
        }

    def __remove_private_var(self, raw_cfg):
        tmp_dict = {}
        for k, v in raw_cfg.__dict__.items():
            if not k[0] == '_':
                tmp_dict[k] = v
        return tmp_dict

    def cnvt_cfg(self, cfg: Configer, target_cfg_type:str, **cnvtr_kwarg):
        assert target_cfg_type in self.dispatcher.keys(), '''Unfortunately, {} config is not supported yet\n
                                                Currently, easy_configer only support : {}'''.format(
                                                    target_cfg_type, self.dispatcher.keys()
                                            )
        return self.dispatcher[target_cfg_type](cfg, **cnvtr_kwarg)

    def argparse_cnvt(self, raw_cfg, parse_arg=True, flatten_args=True):
        args_template = ArgumentParser( description=raw_cfg.get_doc_str() )
        raw_dict = self.__remove_private_var(raw_cfg)
        
        for sec_key, sec_val in raw_dict.items():
            if flatten_args and isinstance(sec_val, dict):
                for key, val in sec_val.items():
                    args_template.add_argument(f"--{key}", type=type(val), default=val)
            else:
                args_template.add_argument(f"--{sec_key}", type=type(sec_val), default=sec_val)

        return args_template.parse_args() if parse_arg else args_template

    def yaml_cnvt(self, raw_cfg):
        raw_dict = self.__remove_private_var(raw_cfg)
        return yaml.dump(raw_dict)

    def omegacfg_cvnt(self, raw_cfg):
        raw_dict = self.__remove_private_var(raw_cfg)
        return OmegaConf.create(raw_dict)
        


import sys
from .utils.Type_Convertor import Type_Convertor

class Argparser:

    # Update the namespace value via commend-line input 
    @staticmethod
    def args_from_cmd(idx_sec_by_dot):   
        '''
            Update the arguments by commend line input string
        '''
        # Note that argument key should begin with - symbol for increasing read-ability!
        def is_arg_key(cmd_str):
            if cmd_str[0] == '-':
                return True
            else:
                return False

        typ_cnvt = Type_Convertor()

        # remove file name from args
        cmd_arg_lst = sys.argv[1:]
        
        # print out helper document string
        if "-h" in cmd_arg_lst:
            cmd_arg_lst.remove("-h")
            print(cfg_dict['_doc_str'])

        sec_ptr, sec_key = None, None
        for idx, item in enumerate(cmd_arg_lst):
            if is_arg_key(item):  # argument flag 
                sec_keys_str = item.split('-')[-1]   # strip - symbol, and get argument key
                sec_ptr, sec_key = idx_sec_by_dot(sec_keys_str)       
            else:                # argument value 
                if idx % 2 == 0:
                    raise RuntimeError('Commendline argument missing the argument key, note that argument key should begin with - symbol!')
                val_str = item
                sec_ptr[sec_key] = typ_cnvt.convert(val_str)
                


import unittest

from easy_config.Configer import Configer
from easy_config.IO_Converter import IO_Converter

import argparse 

class IO_ConverterTestCase(unittest.TestCase):

    def setUp(self):
        self.cfger = Configer()
        self.cfger.cfg_from_ini("./test_cfg_a.ini")
        self.cfger.cfg_from_str(self._build_cfg_text())

    def _build_cfg_text(self):
        return '''
        # Initial config file :
        [Section_test_A]         
        mrg_var_tst = fromBcfg@str

        [Section_test_B]
        fflg = False@bool   # test inline comment in cfg-str
        tflg = True@bool
        # Cell cfg written by Josef-Huang..
        '''
    
    def test_io_convter(self):
        cfg_cnvter = IO_Converter()
        
        argp_cfg = cfg_cnvter.cnvt_cfg(self.cfger, 'argparse')
        ome_cfg = cfg_cnvter.cnvt_cfg(self.cfger, 'omegacfg')
        yaml_cfg = cfg_cnvter.cnvt_cfg(self.cfger, 'yaml')
        
        self.assertEqual(type(argp_cfg), argparse.Namespace)
        self.assertEqual(ome_cfg.Section_test_B.fflg, False)
        self.assertEqual(type(yaml_cfg), str)

if __name__ == '__main__':
    tests = ['test_io_convter']
    suite = unittest.TestSuite(map(IO_ConverterTestCase, tests))
    unittest.main(verbosity=2)


import unittest

from easy_config.Configer import Configer
from get_flag import get_flag_from_ext

class test(object):    
    def __init__(self, num, string):
        self.num = num
        self.name = string

    def num_name(self):
        return self.num * self.name

class ConfigerTestCase(unittest.TestCase):
    
    def setUp(self):
        self.cfgerA = Configer()
        self.cfgerB = Configer()
        self.cfgerC = Configer()

    def _build_cfg_text(self):
        return '''
        # Initial config file :
        [Section_test_A]         
        mrg_var_tst = fromBcfg@str

        [Section_test_B]
        fflg = False@bool   # test inline comment in cfg-str
        tflg = True@bool
        # Cell cfg written by Josef-Huang..
        '''
    
    def test_parse_config(self):
        self.cfgerA.cfg_from_ini("./test_cfg_a.ini")
        raw_cfg_text = self._build_cfg_text()
        self.cfgerB.cfg_from_str(raw_cfg_text)

        # Basic test for config A
        self.assertEqual(type(self.cfgerA.Section_test_A), dict)
        self.assertEqual(self.cfgerA.Section_test_A['sec_var_tst'], "test")
        self.assertEqual(self.cfgerA.Section_test_A['tst_ext_var'], self.cfgerA.flt_var)
        
        # Basic test for config B
        self.assertEqual(type(self.cfgerB.Section_test_B), dict)
        self.assertEqual(self.cfgerB.Section_test_B['fflg'], False)
        self.assertEqual(self.cfgerB.Section_test_A['mrg_var_tst'], 'fromBcfg')

    def test_merge_config(self):
        self.cfgerA.cfg_from_ini("./test_cfg_a.ini")
        raw_cfg_text = self._build_cfg_text()
        self.cfgerB.cfg_from_str(raw_cfg_text)

        self.cfgerA.merge_conf(self.cfgerB, override=True)
        # after merging, secB should in cfgerA, the var should be replaced!!
        self.assertIn('Section_test_B', self.cfgerA.__dict__)
        self.assertEqual(self.cfgerA.Section_test_A['mrg_var_tst'], 'fromBcfg')

    def test_concate_config(self):
        self.cfgerA.cfg_from_ini("./test_cfg_a.ini")
        raw_cfg_text = self._build_cfg_text()
        self.cfgerB.cfg_from_str(raw_cfg_text)

        merge_conf = self.cfgerA.concate_cfg(self.cfgerB, override=True)
        self.assertIn('Section_test_B', merge_conf.__dict__)
        self.assertNotIn('Section_test_B', self.cfgerA.__dict__)
        self.assertNotEqual(self.cfgerA.Section_test_A['mrg_var_tst'], 'fromBcfg')

    def test_self_regist_cls(self):
        self.cfgerC.regist_cnvtor('tst_cls', test)
        self.cfgerC.cfg_from_ini("./test_cfg_reg.ini")
        # test dummy class
        self.assertEqual(type(self.cfgerC.dummy), test)
        self.assertEqual(self.cfgerC.dummy.name, 'josef')
        self.assertEqual(self.cfgerC.Section_test_A['dummy'].num_name(), 3*'ulala')

if __name__ == '__main__':
    tests = ['test_parse_config', 'test_merge_config', 'test_self_regist_cls']
    suite = unittest.TestSuite(map(ConfigerTestCase, tests))
    unittest.main(verbosity=2)